@@ -218,7 +218,7 @@ def skip_first_whitespace(box, skip_stack):
     return None
 
 
-def remove_last_whitespace(context, box):
+def remove_last_whitespace(context, line):
     """Remove in place space characters at the end of a line.
 
     This also reduces the width and position of the inline parents of the
@@ -226,6 +226,7 @@ def remove_last_whitespace(context, box):
 
     """
     ancestors = []
+    box = line
     while isinstance(box, (boxes.LineBox, boxes.InlineBox)):
         ancestors.append(box)
         if not box.children:
@@ -244,25 +245,16 @@ def remove_last_whitespace(context, box):
         assert resume is None
         space_width = box.width - new_box.width
         box.width = new_box.width
-
-        # RTL line, the trailing space is at the left of the box. We have to
-        # translate the box to align the stripped text with the right edge of
-        # the box.
-        if new_box.pango_layout.first_line_direction % 2:
-            box.position_x -= space_width
-            for ancestor in ancestors:
-                ancestor.position_x -= space_width
     else:
         space_width = box.width
         box.width = 0
         box.text = ''
 
-        # RTL line, the textbox with a trailing space is now empty at the left
-        # of the line. We have to translate the line to align it with the right
-        # edge of the box.
-        line = ancestors[0]
-        if line.style['direction'] == 'rtl':
-            line.translate(dx=-space_width, ignore_floats=True)
+    # RTL line, the trailing space is at the left of the box. We have to translate the
+    # box to align the stripped text with the right edge of the box.
+    if box.pango_layout.first_line_direction % 2:
+        for child in line.children:
+            child.translate(dx=-space_width, ignore_floats=True)
 
     for ancestor in ancestors:
         ancestor.width -= space_width
@@ -370,9 +362,12 @@ def atomic_box(context, box, position_x, skip_stack, containing_block,
         if box.is_table_wrapper:
             containing_size = (containing_block.width, containing_block.height)
             table_wrapper_width(context, box, containing_size)
+            width, min_width, max_width = box.width, box.min_width, box.max_width
         box = inline_block_box_layout(
             context, box, position_x, skip_stack, containing_block,
             absolute_boxes, fixed_boxes)
+        if box.is_table_wrapper:
+            box.width, box.min_width, box.max_width = width, min_width, max_width
     else:  # pragma: no cover
         raise TypeError(f'Layout for {type(box).__name__} not handled yet')
     return box
@@ -513,7 +508,7 @@ def split_inline_level(context, box, position_x, max_x, bottom_space,
                 setattr(box, f'margin_{side}', 0)
         new_box, resume_at, _, _, _ = flex_layout(
             context, box, -inf, skip_stack, containing_block, False,
-            absolute_boxes, fixed_boxes)
+            absolute_boxes, fixed_boxes, False)
         preserved_line_break = False
         first_letter = '\u2e80'
         last_letter = '\u2e80'
@@ -1049,8 +1044,7 @@ def inline_box_verticality(box, top_bottom_subtrees, baseline_y):
 
         # the child’s `top` is `child.baseline` above (lower y) its baseline.
         top = child_baseline_y - child.baseline
-        box_types = (
-            boxes.InlineBlockBox, boxes.InlineFlexBox, boxes.InlineGridBox)
+        box_types = (boxes.InlineBlockBox, boxes.InlineFlexBox, boxes.InlineGridBox)
         if isinstance(child, box_types):
             # This also includes table wrappers for inline tables.
             child.translate(dy=top - child.position_y)
@@ -1119,18 +1113,17 @@ def text_align(context, line, available_width, last):
 def justify_line(context, line, extra_width):
     # TODO: We should use a better algorithm here, see
     # https://www.w3.org/TR/css-text-3/#justify-algos
-    nb_spaces = count_spaces(line)
-    if nb_spaces == 0:
-        return
-    add_word_spacing(context, line, extra_width / nb_spaces, 0)
+    if (nb_spaces := count_expandable_spaces(line)):
+        add_word_spacing(context, line, extra_width / nb_spaces, 0)
 
 
-def count_spaces(box):
+def count_expandable_spaces(box):
+    """Count expandable spaces (space and nbsp) for justification."""
     if isinstance(box, boxes.TextBox):
         # TODO: remove trailing spaces correctly
-        return box.text.count(' ')
+        return box.text.count(' ') + box.text.count('\u00a0')
     elif isinstance(box, (boxes.LineBox, boxes.InlineBox)):
-        return sum(count_spaces(child) for child in box.children)
+        return sum(count_expandable_spaces(child) for child in box.children)
     else:
         return 0
 
@@ -1139,7 +1132,7 @@ def add_word_spacing(context, box, justification_spacing, x_advance):
     if isinstance(box, boxes.TextBox):
         box.justification_spacing = justification_spacing
         box.position_x += x_advance
-        nb_spaces = count_spaces(box)
+        nb_spaces = count_expandable_spaces(box)
         if nb_spaces > 0:
             layout = create_layout(
                 box.text, box.style, context, max_width=None,
@@ -1152,7 +1145,10 @@ def add_word_spacing(context, box, justification_spacing, x_advance):
     elif isinstance(box, (boxes.LineBox, boxes.InlineBox)):
         box.position_x += x_advance
         previous_x_advance = x_advance
-        for child in box.children:
+        children = box.children
+        if box.style['direction'] == 'rtl':
+            children = children[::-1]
+        for child in children:
             if child.is_in_normal_flow():
                 x_advance = add_word_spacing(
                     context, child, justification_spacing, x_advance)
