@@ -5,7 +5,7 @@ from math import inf
 
 import pyphen
 
-from .constants import LST_TO_ISO, PANGO_WRAP_MODE
+from .constants import LST_TO_ISO, PANGO_DIRECTION, PANGO_WRAP_MODE
 from .ffi import FROM_UNITS, TO_UNITS, ffi, gobject, pango, pangoft2, unicode_to_char_p
 from .fonts import font_features, get_font_description
 
@@ -76,6 +76,8 @@ class Layout:
             pango.pango_font_map_create_context(font_map),
             gobject.g_object_unref)
         pango.pango_context_set_round_glyph_positions(pango_context, False)
+        pango.pango_context_set_base_dir(
+            pango_context, PANGO_DIRECTION[style['direction']])
 
         if style['font_language_override'] != 'normal':
             lang_p, lang = unicode_to_char_p(LST_TO_ISO.get(
@@ -96,6 +98,7 @@ class Layout:
         self.layout = ffi.gc(
             pango.pango_layout_new(pango_context),
             gobject.g_object_unref)
+        pango.pango_layout_set_auto_dir(self.layout, False)
         pango.pango_layout_set_font_description(self.layout, font_description)
 
         text_decoration = style['text_decoration_line']
@@ -187,13 +190,11 @@ class Layout:
                     pango.pango_layout_set_text(self.layout, text, -1)
 
                 space_spacing = int(word_spacing * TO_UNITS + letter_spacing)
-                position = bytestring.find(b' ')
                 # Pango gives only half of word-spacing on boundaries
                 boundary_positions = (0, len(bytestring) - 1)
-                while position != -1:
-                    factor = 1 + (position in boundary_positions)
-                    add_attr(position, position + 1, factor * space_spacing)
-                    position = bytestring.find(b' ', position + 1)
+                for match in re.finditer(' |\u00a0'.encode(), bytestring):
+                    factor = 1 + (match.start() in boundary_positions)
+                    add_attr(match.start(), match.end(), factor * space_spacing)
 
             if word_breaking:
                 attr = pango.pango_attr_insert_hyphens_new(False)
@@ -213,7 +214,7 @@ class Layout:
             layout.set_text(' ' * self.style['tab_size'])
             line, _ = layout.get_first_line()
             width, _ = line_size(line, self.style)
-            width = int(round(width))
+            width = round(width)
         else:
             width = int(self.style['tab_size'].value)
         # 0 is not handled correctly by Pango
